import subprocess
from typing import Any, Callable
from urllib.parse import ParseResult, urlunparse, urldefrag, unquote
import logging
from dataclasses import dataclass
import os
import json
import socket

from microkodi.process import nonblocking_run
from microkodi.config import Config
from microkodi.programs import PlayerInfo, Program
from microkodi.repository import I

class MpvConfig:
    # Mapping of file extensions to a potential transformer function
    _ext_map: dict[str, Callable[[ParseResult], tuple[list[str], str]]]

    def __init__(self):
        self._ext_map = {}

    def register_ext_transformer(self, ext: str, transformer: Callable[[ParseResult], tuple[list[str], str]]):
        logging.getLogger("MpvConfig").debug("Registering transformer for .%s URLs", ext)
        self._ext_map[ext] = transformer

    def get_ext_transformer(self, ext: str) -> Callable[[ParseResult], tuple[list[str], str]] | None:
        return self._ext_map.get(ext)
I.register("MpvConfig", MpvConfig())

class MpvProgram(Program):
    _process: subprocess.Popen

    _mpv_bin_path: str
    _mpv_extra_args: list[str]
    _mpv_ipc_sock: str

    def __init__(self):
        super().__init__()

        config: Config = I.get("Config")
        self._process = None
        self._mpv_bin_path = config.mpv
        self._mpv_extra_args = config.mpv_args
        self._mpv_ipc_sock = config.mpv_ipc_sock

    def __mpv_command(self, args: list[str]) -> dict[str, Any]:
        ipc_logger = logging.getLogger("mpv-ipc")
        command = json.dumps({
            "command": args,
        }) + "\n"
        ipc_logger.debug("-> %s", command)
        with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as client:
            try:
                client.connect(self._mpv_ipc_sock)
            except ConnectionRefusedError:
                ipc_logger.warn("Connection refused")
                return {}
            client.send(command.encode("utf-8"))

            resp_raw = client.recv(4096).decode()
            ipc_logger.debug("<- %s", resp_raw)
            return json.loads(resp_raw)


    def __read_prop(self, prop: str, default: Any) -> Any:
        return self.__mpv_command(["get_property", prop]).get("data", default)

    def is_active(self) -> bool:
        return self._process is not None and self._process.returncode is None

    def is_playing(self) -> bool:
        return not self.__read_prop("pause", False)

    def resume(self):
        self.__mpv_command(["set_property", "pause", False])

    def play(self, url: ParseResult):
        self.stop("mpvplayer.play")

        # Allow preprocessing the URL
        last_segment = url.path.split("/")[-1]
        final_url = ""
        extra_args = []
        if "." in last_segment:
            *_, ext = last_segment.split(".")
            transformer = I.get("MpvConfig").get_ext_transformer(ext)
            if transformer is not None:
                self.logger.info("Transforming URL due to extension: %s", ext)
                extra_args, final_url = transformer(url)
                self.logger.info("New url: %s", final_url)
            else:
                final_url = urlunparse(url)
        else:
            final_url = urlunparse(url)

        # Start mpv
        cmdline = [
            self._mpv_bin_path,
            f"--input-ipc-server={self._mpv_ipc_sock}",
            "--fs",
            "--quiet",
            *self._mpv_extra_args,
            *extra_args,
            final_url,
        ]
        self.logger.debug("Cmdline: %s", " ".join(cmdline))
        self._process = nonblocking_run(cmdline, self._when_mpv_exit)

    def _when_mpv_exit(self):
        self.logger.info("MPV has exited")
        I.get("DataBridge").set_loading(False)

        if self._process.returncode != 0:
            I.get("DataBridge").notification.emit("mpv exited with an error")
        self._process = None

    def pause(self):
        self.__mpv_command(["set_property", "pause", True])

    def stop(self, src):
        if self.is_active():
            self._process.terminate()

    def exit(self):
        self.stop("mpvplayer.exit")

    def get_player_info(self) -> PlayerInfo:
        if not os.path.exists(self._mpv_ipc_sock):
            return PlayerInfo(
                runtime=0,
                position=0,
                title="",
                playing=False,
            )

        return PlayerInfo(
            runtime=int(self.__read_prop("duration", "0")),
            position=int(self.__read_prop("playback-time", "0")),
            title=self.__read_prop("media-title", ""),
            playing=not self.__read_prop("pause", False),
        )